/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi;

import org.javafmi.kernel.FmuProxy;
import org.javafmi.kernel.ProxyFactory;
import org.javafmi.kernel.Status;
import org.javafmi.kernel.operatingsystem.FmuFile;
import org.javafmi.modeldescription.ModelDescription;

import java.io.File;
import java.util.List;

import static org.javafmi.kernel.Status.OK;

public class Simulation {

    private static final int DEFAULT_TIMEOUT = 1000;
    private VariableWriter variableWriter;
    private VariableReader variableReader;
    private FmuProxy fmuProxy;
    private double simulationTime;
    private double startTime;
    private double stopTime;

    protected Simulation() {
    }

    public Simulation(String fmuPath) {
        this(fmuPath, DEFAULT_TIMEOUT);
    }

    public Simulation(String fmuPath, double timeout) {
        this.fmuProxy = ProxyFactory.createProxyFor(new FmuFile(new File(fmuPath)));
        this.fmuProxy.instantiate(timeout);
        this.variableWriter = new VariableWriter(fmuProxy, fmuProxy.getModelDescription());
        this.variableReader = new VariableReader(fmuProxy, fmuProxy.getModelDescription());
    }

    protected static FmuProxy getProxy(Simulation simulation) {
        return simulation.fmuProxy;
    }

    public ModelDescription getModelDescription() {
        return fmuProxy.getModelDescription();
    }

    public Status init(double startTime) {
        return init(startTime, 2 ^ 32);
    }

    public Status init(double startTime, double stopTime) {
        simulationTime = startTime;
        this.startTime = startTime;
        this.stopTime = stopTime;
        return fmuProxy.initialize(startTime, stopTime);
    }

    public Status doStep(double stepSize) {
        Status doStepStatus = fmuProxy.doStep(simulationTime, stepSize);
        if(doStepStatus == OK) simulationTime += stepSize;
        return doStepStatus;
    }

    public Status cancelStep() {
        return fmuProxy.cancelStep();
    }

    public Status terminate() {
        Status terminateStatus = fmuProxy.terminate();
        fmuProxy.freeInstance();
        return terminateStatus;
    }

    public Status reset() {
        Status resetStatus = fmuProxy.reset();
        if (resetStatus.equals(OK)) simulationTime = startTime;
        return resetStatus;
    }

    public Variable readVariable(String variableName) {
        return variableReader.read(variableName);
    }

    public Status writeVariable(String name, Object value) {
        return variableWriter.write(name, value);
    }

    public boolean isTerminated() {
        return fmuProxy.isTerminated() || simulationTime >= stopTime;
    }

    public double getSimulationTime() {
        return simulationTime;
    }

    public void setSimulationTolerance(double simulationTolerance) {
        fmuProxy.setTolerance(simulationTolerance);
    }

    public void writeVariables(List<Variable> variablesToWrite) {
        variableWriter.write(variablesToWrite.toArray(new Variable[variablesToWrite.size()]));
    }

    public Variable[] readVariables(String[] variableNames) {
        return variableReader.read(variableNames);
    }
}
