/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.kernel.operatingsystem;

import org.javafmi.exceptions.JavaFmiException;

import java.io.File;

import static java.lang.System.getProperty;

public class FmuFile {

    public static final String BINARIES_FOLDER = "binaries";
    public static final String MAC_OS_FOLDER = "darwin";
    public static final String WINDOWS_FOLDER = "win";
    public static final String LINUX_FOLDER = "linux";
    public static final String MAC_OS_LIBRARY_EXTENSION = ".dylib";
    public static final String WINDOWS_LIBRARY_EXTENSION = ".dll";
    public static final String LINUX_LIBRARY_EXTENSION = ".so";
    public static final String ARCH_64 = "64";
    public static final String ARCH_32 = "32";
    public static final String MISSING_SYSTEM_MESSAGE = "JavaFMI is not prepared to your system";
    private static final String MODEL_DESCRIPTION_FILENAME = "modelDescription.xml";
    private static final String RESOURCES_FOLDER = "resources";
    private final File fmuFile;
    private File workingDirectory;
    private UnZipper unZipper;
    private File resourcesDirectory;

    public FmuFile(String filePath) {
        this(new File(filePath));
    }

    public FmuFile(File fmuFile) {
        this.fmuFile = fmuFile;
        this.unZipper = new UnZipper(fmuFile);
        this.workingDirectory = createTemporal();
    }

    private File createTemporal() {
        String systemTempDirectory = System.getProperty("java.io.tmpdir");
        File tempFolder = new File(PathBuilder.buildPath(systemTempDirectory, "fmu_" + System.currentTimeMillis() + fmuFile.getName().replace(".fmu", "")));
        tempFolder.mkdirs();
        tempFolder.deleteOnExit();
        return tempFolder;
    }

    public void deleteTemporal() {
        deleteAll(workingDirectory);
    }

    public File getModelDescriptionFile() {
        return unZipper.unzip(MODEL_DESCRIPTION_FILENAME, workingDirectory);
    }

    public String getLibraryPath() {
        return getLibraryFile().getAbsolutePath().replaceAll("\\\\", "/");
    }

    public File getLibraryFile() {
        String libraryPathInsideZip = PathBuilder.buildPath(new String[]{BINARIES_FOLDER, getLibraryFolder() + getSystemArchitecture()});
        return unZipper.unzipFileWithExtension(libraryPathInsideZip, getLibraryExtension(), workingDirectory);
    }

    private String getLibraryFolder() {
        if(systemIsMacOS(getSystemName())) return MAC_OS_FOLDER;
        else if(systemIsWindows(getSystemName())) return WINDOWS_FOLDER;
        else if(systemIsLinux(getSystemName())) return LINUX_FOLDER;
        throw new JavaFmiException(MISSING_SYSTEM_MESSAGE);

    }


    private String getSystemArchitecture() {
        return is64bitsArchitecture(getProperty("os.arch").toLowerCase()) ? ARCH_64 : ARCH_32;
    }

    private String getLibraryExtension() {
        if(systemIsMacOS(getSystemName())) return MAC_OS_LIBRARY_EXTENSION;
        else if(systemIsWindows(getSystemName())) return WINDOWS_LIBRARY_EXTENSION;
        else if(systemIsLinux(getSystemName())) return LINUX_LIBRARY_EXTENSION;
        throw new JavaFmiException(MISSING_SYSTEM_MESSAGE);
    }

    private String getSystemName() {
        return getProperty("os.name").toLowerCase();
    }

    private boolean systemIsMacOS(String operatingSystem) {
        return operatingSystem.startsWith("mac");
    }

    private boolean systemIsWindows(String operatingSystemName) {
        return operatingSystemName.startsWith("win");
    }

    private boolean systemIsLinux(String operatingSystemName) {
        return operatingSystemName.startsWith("linux");
    }

    private boolean is64bitsArchitecture(String architecture) {
        return architecture.contains(ARCH_64);
    }

    public String getWorkingDirectoryPath() {
        return workingDirectory.getAbsolutePath();
    }

    private void deleteAll(File file) {
        if(file.isDirectory())
            for(File child : file.listFiles())
                deleteAll(child);
        file.delete();
    }


    public String getResourcesDirectoryPath() {
        return getResourcesDirectory().getPath().replaceAll("\\\\", "/");
    }

    public File getResourcesDirectory() {
        try {
            return (resourcesDirectory != null) ?
                    resourcesDirectory :
                    (resourcesDirectory = unZipper.unzipAllFilesInDirectory("resources", workingDirectory));
        } catch(EntryNotFoundException e) {
            return FileUtils.createFolder(workingDirectory, "resources");
        }
    }
}
