/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.kernel.operatingsystem;

import org.javafmi.exceptions.JavaFmiException;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import static org.javafmi.kernel.operatingsystem.FileUtils.*;

public class UnZipper {

    private ZipFile zip;

    public UnZipper(File file) {
        if(!file.exists()) throw new JavaFmiException(String.format("File %s does not exist", file.getAbsolutePath()));
        zip = openZipForFile(file);
    }

    private ZipFile openZipForFile(File file) {
        try {
            return new ZipFile(file);
        } catch(IOException e) {
            throw new RuntimeException(e);
        }
    }

    public File unzip(String fileName, File outFolder) {
        if(!outFolder.exists()) createFolder(outFolder);
        return extractEntryToFile(zip.getEntry(fileName), createFile(outFolder, fileName));
    }

    public File unzipFileWithExtension(String entryPath, String extension, File outPath) {
        Enumeration<? extends ZipEntry> entries = zip.entries();
        while(entries.hasMoreElements()) {
            ZipEntry entry = entries.nextElement();
            if(entry.getName().replace("\\", "/").startsWith(entryPath) && entry.getName().endsWith(extension))
                return unzip(entry.getName(), outPath);
        }
        throw new RuntimeException(String.format("There are no entries in %s with extension %s", entryPath, extension));
    }

    public File unzipAllFilesInDirectory(String directory, File outPath) {
        String entryToSearchName = directory + '/';
        ZipEntry entryToUnzip = zip.getEntry(entryToSearchName);
        if(entryToUnzip == null) throw new EntryNotFoundException("entry not found: " + entryToSearchName);
        if(!entryToUnzip.isDirectory()) throw new JavaFmiException(directory + " is not a directory");
        File extractedDirectory = createFolder(outPath, directory);
        Enumeration<? extends ZipEntry> entries = zip.entries();
        while(entries.hasMoreElements()) {
            ZipEntry entry = entries.nextElement();
            if(entry.getName().startsWith(directory) && !entry.getName().equals(directory + "/")) {
                extractEntryToFile(entry, createFile(outPath, entry.getName()));
            }
        }
        return extractedDirectory;
    }


    private File extractEntryToFile(ZipEntry entry, File outFile) {
        try {
            BufferedInputStream source = bufferedInputStream(zip.getInputStream(entry));
            BufferedOutputStream target = bufferedOutputStream(fileOutputStream(outFile));
            copyFromTo(source, target);
            close(source);
            close(target);
            return outFile;
        } catch(Exception e) {
            throw new RuntimeException("Impossible to extract " + outFile.getName(), e);
        }
    }

    public void unzipAll(File outputDirectory) {
        try {
            createFolder(outputDirectory);
            Enumeration zipFileEntries = zip.entries();
            while(zipFileEntries.hasMoreElements()) {
                ZipEntry entry = (ZipEntry) zipFileEntries.nextElement();
                if(!entry.isDirectory()) {
                    BufferedInputStream source = bufferedInputStream(zip.getInputStream(entry));
                    BufferedOutputStream target = bufferedOutputStream(fileOutputStream(createFile(outputDirectory, entry.getName())));
                    copyFromTo(source, target);
                    flush(target);
                    close(source);
                    close(target);
                } else {
                    createFolder(outputDirectory, entry.getName());
                }
            }
        } catch(IOException e) {
            e.printStackTrace();
        }
    }
}
