/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.kernel.v1;

import com.sun.jna.Function;
import com.sun.jna.Memory;
import com.sun.jna.NativeLibrary;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.PointerByReference;
import org.javafmi.exceptions.MethodExecutionException;
import org.javafmi.kernel.FmuProxyState;
import org.javafmi.kernel.SharedMemoryAllocator;
import org.javafmi.kernel.Status;
import org.javafmi.kernel.operatingsystem.FmuFile;
import org.javafmi.kernel.operatingsystem.NativeBoolean;
import org.javafmi.kernel.v1.nativeaccess.CallbackFunctionsStructure;
import org.javafmi.kernel.v1.nativeaccess.FmuFunctionSuffix;
import org.javafmi.kernel.v1.nativeaccess.Size_T;
import org.javafmi.modeldescription.FmiVersion;
import org.javafmi.modeldescription.ModelDescriptionDeserializer;
import org.javafmi.modeldescription.ScalarVariable;
import org.javafmi.modeldescription.v1.EnumerationType;
import org.javafmi.modeldescription.v1.ModelDescription;

import java.nio.DoubleBuffer;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;

import static org.javafmi.kernel.FmuProxyState.*;
import static org.javafmi.kernel.Status.*;
import static org.javafmi.kernel.v1.nativeaccess.FmuFunctionSuffix.*;

public class FmuProxy implements org.javafmi.kernel.FmuProxy {

    private static final String MIME_TYPE = "application/x-fmu-sharedlibrary";
    private String proxyID;
    private String modelIdentifier;
    private ModelDescription modelDescription;
    private FmuFile fmuFile;
    private NativeLibrary libraryInstance;
    private boolean isInstanceVisible;
    private boolean isInstanceInteractive;
    private boolean isDoingLogging;
    private boolean isStopTimeDefined;
    private double startTime;
    private double stopTime;
    private double timeout;
    private Pointer instantiatedSlave;
    private Map<String, Function> nativeFunctions;
    private FmuProxyState fmuProxyState;
    private Size_T sizeT;
    private CallbackFunctionsStructure.ByValue callbacks;
    private SharedMemoryAllocator sharedMemoryAllocator;

    public FmuProxy(FmuFile fmuFile) {
        this.fmuFile = fmuFile;
        setUp();
    }

    @Override
    public void instantiate(double timeout) {
        this.timeout = timeout;
        this.libraryInstance = NativeLibrary.getInstance(this.fmuFile.getLibraryPath());
        this.nativeFunctions = getNativeFunctions();
        this.instantiatedSlave = getPointerToLibrary();
        this.sizeT = new Size_T(1);
        this.fmuProxyState = INSTANTIATED;
    }

    public void setUp() {
        this.callbacks = new CallbackFunctionsStructure.ByValue();
        this.sharedMemoryAllocator = new SharedMemoryAllocator();
        this.modelDescription = (ModelDescription) ModelDescriptionDeserializer.deserialize(ModelDescription.class, fmuFile.getModelDescriptionFile());
        modelDescription.buildAttributes();
        this.modelIdentifier = modelDescription.getModelIdentifier();
        this.proxyID = modelIdentifier + "_" + UUID.randomUUID().toString();
        this.isInstanceVisible = false;
        this.isInstanceInteractive = false;
        this.isDoingLogging = false;
        this.isStopTimeDefined = true;
    }

    private Pointer getPointerToLibrary() {
        Pointer pointerToLibrary = (Pointer) nativeFunctions.get(INSTANTIATE).invoke(Pointer.class, getParametersForInstantiate());
        if(pointerToLibrary == Pointer.NULL)
            throw new MethodExecutionException("Instantiate: " + modelIdentifier);
        return pointerToLibrary;
    }

    public Map<String, Function> getNativeFunctions() {
        LinkedHashMap<String, Function> functions = new LinkedHashMap<>();
        for(String suffix : FmuFunctionSuffix.getFunctionsSuffix())
            functions.put(suffix, getFunction(suffix));
        return functions;
    }

    @Override
    public Status initialize(double startTime, double stopTime) {
        if(fmuProxyState.equals(INITED)) return WARNING;
        this.startTime = startTime;
        this.stopTime = stopTime;
        Status initStatus = translateStatus((Integer) nativeFunctions.get(INITIALIZE).invoke(Integer.class, getParametersForInitialize()));
        if(initStatus != Status.OK) {
            fmuProxyState = TERMINATED;
            throw new MethodExecutionException("init: " + modelIdentifier);
        }
        fmuProxyState = INITED;
        return initStatus;
    }

    @Override
    public Status terminate() {
        Status terminateStatus = translateStatus((Integer) nativeFunctions.get(TERMINATE).invoke(Integer.class, new Object[]{instantiatedSlave}));
        if(terminateStatus.equals(OK)) fmuProxyState = TERMINATED;
        else fmuProxyState = INITED;
        return terminateStatus;
    }

    @Override
    public Status reset() {
        Status resetStatus = translateStatus((Integer) nativeFunctions.get(RESET).invoke(Integer.class, new Object[]{instantiatedSlave}));
        if(resetStatus.equals(Status.OK)) {
            fmuProxyState = INSTANTIATED;
        }
        return resetStatus;
    }

    @Override
    public org.javafmi.modeldescription.ModelDescription getModelDescription() {
        return modelDescription;
    }

    @Override
    public void freeInstance() {
        nativeFunctions.get(FREE).invoke(new Object[]{instantiatedSlave});
        libraryInstance.dispose();
        fmuFile.deleteTemporal();
    }

    @Override
    public Status doStep(double communicationPoint, double stepSize) {
        return translateStatus(nativeFunctions.get(DO_STEP).invokeInt(getParametersForDoStep(stepSize, communicationPoint)));
    }

    @Override
    public Status cancelStep() {
        return translateStatus((Integer) nativeFunctions.get(CANCEL_STEP).invoke(Integer.class, new Object[]{instantiatedSlave}));
    }

    @Override
    public Object getEnumeration(ScalarVariable modelVariable) {
        EnumerationType enumeration = (EnumerationType) modelVariable.getType();
        Integer enumerationIndex;
        EnumerationType enumerationType = (EnumerationType) modelDescription.getTypeFromTypeDefinition(enumeration.getDeclaredType());
        if(modelVariable.getVariability().equalsIgnoreCase("constant")) {
            enumerationIndex = enumeration.getStart();
            return enumerationType.getItems().get(enumerationIndex - 1).getName();
        }
        enumerationIndex = getInteger(modelVariable.getValueReference());
        return enumerationType.getItems().get(enumerationIndex - 1).getName();
    }

    @Override
    public double[] getReal(int... valueReferences) {
        DoubleBuffer buffer = DoubleBuffer.allocate(valueReferences.length);
        nativeFunctions.get(GET_REAL).invokeInt(new Object[]{instantiatedSlave, valueReferences, valueReferences.length, buffer});
        return buffer.array();
    }

    @Override
    public int getInteger(Integer valueReference) {
        nativeFunctions.get(GET_INTEGER).invoke(getParametersForFmuGetterSetter(valueReference, sharedMemoryAllocator.getSharedInteger()));
        return sharedMemoryAllocator.getSharedInteger().get(0);
    }

    @Override
    public boolean getBoolean(Integer valueReference) {
        nativeFunctions.get(GET_BOOLEAN).invoke(getParametersForFmuGetterSetter(valueReference, sharedMemoryAllocator.getSharedByte()));
        return NativeBoolean.getBooleanFor(sharedMemoryAllocator.getSharedByte().get(0));
    }

    @Override
    public String getString(Integer valueReference) {
        PointerByReference outPointerByReference = new PointerByReference();
        nativeFunctions.get(GET_STRING).invoke(getParametersForFmuGetterSetter(valueReference, outPointerByReference));
        Pointer stringPointer = outPointerByReference.getValue();
        return (stringPointer != null) ? stringPointer.getString(0) : "";
    }

    @Override
    public Status setReal(int[] valueReferences, double[] values) {
        return translateStatus(nativeFunctions.get(SET_REAL).invokeInt(new Object[]{instantiatedSlave, valueReferences, valueReferences.length, values}));
    }

    @Override
    public Status setInteger(Integer valueReference, Integer newValue) {
        sharedMemoryAllocator.getSharedInteger().put(0, newValue);
        return translateStatus((Integer) nativeFunctions.get(SET_INTEGER).invoke(Integer.class, getParametersForFmuGetterSetter(valueReference, sharedMemoryAllocator.getSharedInteger())));
    }

    @Override
    public Status setBoolean(Integer valueReference, Boolean newValue) {
        sharedMemoryAllocator.getSharedByte().put(0, NativeBoolean.getValueFor(newValue));
        return translateStatus(nativeFunctions.get(SET_BOOLEAN).invokeInt(getParametersForFmuGetterSetter(valueReference, sharedMemoryAllocator.getSharedByte())));
    }

    @Override
    public Status setString(Integer valueReference, String newValue) {
        PointerByReference pointerByReference = new PointerByReference();
        Pointer reference = new Memory(newValue.length() + 1).share(0);
        reference.setString(0, newValue);
        pointerByReference.setValue(reference);
        return translateStatus(nativeFunctions.get(SET_STRING).invokeInt(getParametersForFmuGetterSetter(valueReference, pointerByReference)));
    }

    public boolean isDoingLogging() {
        return isDoingLogging;
    }

    public Status setIsDoingLogging(boolean isDoingLogging) {
        this.isDoingLogging = isDoingLogging;
        return translateStatus((Integer) nativeFunctions.get(SET_DEBUG).invoke(Integer.class, getParametersForSetDebugLogging()));
    }

    @Override
    public boolean isTerminated() {
        return fmuProxyState.equals(TERMINATED);
    }

    @Override
    public void setTolerance(double simulationTolerance) {
        throw new org.javafmi.exceptions.UnsupportedOperationException("Can't set the tolerance in FMI v1 CoSimulation");
    }

    private Function getFunction(String suffix) {
        return libraryInstance.getFunction(modelIdentifier + suffix);
    }

    private Object[] getParametersForInstantiate() {
        return new Object[]{proxyID, modelDescription.getGuid(), fmuFile.getWorkingDirectoryPath(), MIME_TYPE, timeout, NativeBoolean.getValueFor(isInstanceVisible), NativeBoolean.getValueFor(isInstanceInteractive), callbacks, NativeBoolean.getValueFor(isDoingLogging)};
    }

    private Object[] getParametersForInitialize() {
        return new Object[]{instantiatedSlave, startTime, NativeBoolean.getValueFor(isStopTimeDefined), stopTime};
    }

    private Object[] getParametersForDoStep(double stepSize, double communicationPoint) {
        return new Object[]{instantiatedSlave, communicationPoint, stepSize, (byte) 1};
    }

    private Object[] getParametersForSetDebugLogging() {
        return new Object[]{instantiatedSlave, NativeBoolean.getValueFor(isDoingLogging)};
    }

    public Object[] getParametersForFmuGetterSetter(Integer valueReference, Object outBuffer) {
        return new Object[]{instantiatedSlave, sharedMemoryAllocator.getSharedValueReference().put(0, valueReference), sizeT, outBuffer};
    }

    @Override
    public FmiVersion getVersion() {
        return FmiVersion.ONE;
    }
}
