/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.kernel.v2;

import org.javafmi.exceptions.MethodExecutionException;
import org.javafmi.exceptions.UnsupportedOperationException;
import org.javafmi.kernel.FmuProxyState;
import org.javafmi.kernel.State;
import org.javafmi.kernel.Status;
import org.javafmi.kernel.operatingsystem.FmuFile;
import org.javafmi.kernel.v2.nativeaccess.CallbackFunctionsByReference;
import org.javafmi.kernel.v2.nativeaccess.FmiV2;
import org.javafmi.kernel.v2.nativeaccess.FmiV2Interface;
import org.javafmi.modeldescription.FmiVersion;
import org.javafmi.modeldescription.ModelDescriptionDeserializer;
import org.javafmi.modeldescription.ScalarVariable;
import org.javafmi.modeldescription.v2.Capabilities;
import org.javafmi.modeldescription.v2.EnumerationType;
import org.javafmi.modeldescription.v2.ModelDescription;

import java.util.ArrayList;
import java.util.List;

import static org.javafmi.kernel.FmuProxyState.*;
import static org.javafmi.kernel.Status.*;
import static org.javafmi.modeldescription.v2.Capabilities.CAN_GET_AND_SET_FMU_STATE;
import static org.javafmi.modeldescription.v2.Capabilities.PROVIDES_DIRECTIONAL_DERIVATIVE;

public class FmuProxy implements org.javafmi.kernel.FmuProxy {

    private FmuProxyState fmuProxyState;
    private ModelDescription modelDescription;
    private FmiV2 library;
    private FmuFile fmuFile;
    private PendingVariablesList pendingVariables;
    private double simulationTolerance;
    private Capabilities modelDescriptionCapabilities;

    public FmuProxy(FmuFile fmuFile) {
        this.modelDescription = (ModelDescription) ModelDescriptionDeserializer.deserialize(ModelDescription.class, fmuFile.getModelDescriptionFile());
        this.modelDescription.buildAttributes();
        this.modelDescriptionCapabilities = modelDescription.getCapabilities();
        this.fmuFile = fmuFile;
        this.pendingVariables = new PendingVariablesList();
        this.simulationTolerance = modelDescription.getDefaultTolerance();
    }

    @Override
    public FmiVersion getVersion() {
        return FmiVersion.TWO;
    }

    @Override
    public void instantiate(double timeout) {
        this.library = new FmiV2(fmuFile.getLibraryPath());
        System.out.println(fmuFile.getResourcesDirectoryPath());
        List<Object> onMemoryAllocatedObjects = new ArrayList<>();
        this.library.instantiate(
                modelDescription.getModelName(),
                FmiV2Interface.FmiType.fmiCoSimulation,
                modelDescription.getGuid(),
                "file://" + fmuFile.getResourcesDirectoryPath(),
                new CallbackFunctionsByReference(onMemoryAllocatedObjects),
                false,
                false);
        this.fmuProxyState = INSTANTIATED;
    }

    @Override
    public void freeInstance() {
        library.free();
        fmuFile.deleteTemporal();
    }

    public Status setDebugLogging(boolean loggingOn, String[] categories) {
        return Status.translateStatus(library.setDebugLogging(loggingOn, categories));
    }

    @Override
    public Status initialize(double startTime, double stopTime) {
        if (fmuProxyState.equals(INITED)) return WARNING;
        if (ERROR == setUpExperiment(true, simulationTolerance, startTime, false, stopTime))
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        if (ERROR == enterInitializationMode())
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        if (ERROR == setPendingVariables())
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        if (ERROR == exitInitializationMode())
            throw new MethodExecutionException("fmi2SetupExperiment: " + modelDescription.getModelIdentifier());
        return OK;
    }

    private Status setPendingVariables() {
        return pendingVariables.setAll(this);
    }

    public Status setUpExperiment(boolean toleranceDefined, double tolerance, double startTime, boolean stopTimeDefined, double stopTime) {
        return Status.translateStatus(library.setUpExperiment(toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime));
    }

    public Status enterInitializationMode() {
        Status enterInitStatus = Status.translateStatus(library.enterInitializationMode());
        if (enterInitStatus.equals(OK)) fmuProxyState = INITTING;
        return enterInitStatus;
    }

    public Status exitInitializationMode() {
        Status exitInitStatus = Status.translateStatus(library.exitInitializationMode());
        if (exitInitStatus == OK) fmuProxyState = INITED;
        return exitInitStatus;
    }

    @Override
    public Status terminate() {
        Status terminateStatus = Status.translateStatus(library.terminate());
        if (terminateStatus.equals(OK)) fmuProxyState = TERMINATED;
        else fmuProxyState = INITED;
        return terminateStatus;
    }

    @Override
    public Status reset() {
        Status resetStatus = Status.translateStatus(library.reset());
        if (resetStatus.equals(Status.OK)) fmuProxyState = INSTANTIATED;
        return resetStatus;
    }

    @Override
    public org.javafmi.modeldescription.ModelDescription getModelDescription() {
        return modelDescription;
    }

    public Status setRealInputDerivatives(int[] valueReferences, int[] orders, double[] values) {
        return Status.translateStatus(library.setRealInputDerivatives(valueReferences, orders, values));
    }

    public double[] getRealOutputDerivatives(int[] valueReferences, int[] orders) {
        return library.getRealOutputDerivatives(valueReferences, orders);
    }

    @Override
    public double[] getReal(int... valueReference) {
        return library.getReal(valueReference);
    }

    @Override
    public int getInteger(Integer valueReference) {
        return 0;
    }

    @Override
    public boolean getBoolean(Integer valueReference) {
        return false;
    }

    @Override
    public String getString(Integer valueReference) {
        return "empty";
    }

    @Override
    public Object getEnumeration(ScalarVariable modelVariable) {
        EnumerationType enumeration = (EnumerationType) modelVariable.getType();
        Integer enumerationIndex;
        EnumerationType enumerationType = (EnumerationType) modelDescription.getTypeFromTypeDefinition(enumeration.getDeclaredType());
        if (modelVariable.getVariability().equalsIgnoreCase("constant")) {
            enumerationIndex = enumeration.getStart();
            return enumerationType.getItems().get(enumerationIndex - 1).getName();
        }
        enumerationIndex = getInteger(modelVariable.getValueReference());
        return enumerationType.getItems().get(enumerationIndex - 1).getName();
    }

    @Override
    public Status setReal(int[] valueReferences, double[] doubleValues) {
        if ((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            for (int i = 0; i < valueReferences.length; i++) {
                pendingVariables.addReal(valueReferences[i], doubleValues[i]);
            }
            return OK;
        }
        return Status.translateStatus(library.setReal(valueReferences, doubleValues));
    }

    @Override
    public Status setInteger(Integer valueReference, Integer intValue) {
        if ((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addInteger(valueReference, intValue);
            return OK;
        }
        return Status.translateStatus(library.setInteger(new int[]{valueReference}, new int[]{intValue}));
    }

    @Override
    public Status setBoolean(Integer valueReference, Boolean booleanValue) {
        if ((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addBoolean(valueReference, booleanValue);
            return OK;
        }
        return Status.translateStatus(library.setBoolean(new int[]{valueReference}, new boolean[]{booleanValue}));
    }

    @Override
    public Status setString(Integer valueReference, String stringValue) {
        if ((fmuProxyState.equals(FmuProxyState.INSTANTIATED))) {
            pendingVariables.addString(valueReference, stringValue);
            return OK;
        }
        return Status.translateStatus(library.setString(new int[]{valueReference}, new String[]{stringValue}));
    }

    public State getState(State state) {
        if (!modelDescriptionCapabilities.check(CAN_GET_AND_SET_FMU_STATE))
            throw new UnsupportedOperationException("getState: " + modelDescription.getModelIdentifier());
        return state.setState(library.getFmuState(state.getState()));
    }

    public Status setState(State state) {
        if (!modelDescriptionCapabilities.check(CAN_GET_AND_SET_FMU_STATE))
            throw new UnsupportedOperationException("setState: " + modelDescription.getModelIdentifier());
        return Status.translateStatus(library.setFmuState(state.getState()));
    }

    public Status freeState(State state) {
        if (!modelDescriptionCapabilities.check(CAN_GET_AND_SET_FMU_STATE))
            throw new UnsupportedOperationException("freeState: " + modelDescription.getModelIdentifier());
        return Status.translateStatus(library.freeFmuState(state.getState()));
    }

    public int getStateSize(State state) {
        if (!modelDescriptionCapabilities.check(CAN_GET_AND_SET_FMU_STATE) || !modelDescriptionCapabilities.check(Capabilities.CAN_SERIALIZE_FMU_STATE))
            throw new UnsupportedOperationException("getStateSize: " + modelDescription.getModelIdentifier());
        return library.serializedFmuStateSize(state.getState());
    }

    public byte[] serializeState(State state) {
        return library.serializeFmuState(state.getState());
    }

    public State deserializeState(byte[] serializedState) {
        return State.newStateFrom(library.deSerializeFMUState(serializedState));
    }

    public double[] getDirectionalDerivative(int[] unknownVariablesValueReferences, int[] knownVariablesValueReferences, double[] knownDifferential) {
        if (!modelDescriptionCapabilities.check(PROVIDES_DIRECTIONAL_DERIVATIVE))
            throw new UnsupportedOperationException("getDirectionalDerivative: " + modelDescription.getModelIdentifier());
        return library.getDirectionalDerivative(unknownVariablesValueReferences, knownVariablesValueReferences, knownDifferential);
    }

    @Override
    public Status doStep(double communicationPoint, double stepSize) {
        return Status.translateStatus(library.doStep(communicationPoint, stepSize, false));
    }

    @Override
    public Status cancelStep() {
        return Status.translateStatus(library.cancelStep());
    }

    @Override
    public boolean isTerminated() {
        return fmuProxyState.equals(TERMINATED);
    }

    @Override
    public void setTolerance(double simulationTolerance) {
        this.simulationTolerance = simulationTolerance;
    }

    public Capabilities getCapabilities() {
        return modelDescriptionCapabilities;
    }
}
