/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro
 Maria del Carmen Sánchez Medrano

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
 package org.javafmi.kernel.v2.nativeaccess;

import com.sun.jna.Memory;
import com.sun.jna.Native;
import com.sun.jna.Pointer;
import com.sun.jna.ptr.PointerByReference;
import org.javafmi.exceptions.MethodExecutionException;
import org.javafmi.kernel.operatingsystem.NativeBoolean;

public class FmiV2 {

    private final FmiV2Interface fmiV2Interface;
    private Pointer component;

    public FmiV2(String libraryPath) {
        this.fmiV2Interface = (FmiV2Interface) Native.loadLibrary(libraryPath, FmiV2Interface.class);
    }

    public void instantiate(String modelName, int fmiType, String guID, String resourcesLocation, CallbackFunctionsByReference callbacks, boolean visible, boolean loggingOn) {
        component = fmiV2Interface.fmi2Instantiate(
                modelName,
                fmiType,
                guID,
                resourcesLocation,
                callbacks,
                visible,
                loggingOn
        );
        if (component == null)
            throw new MethodExecutionException("Impossible to instantiate " + modelName + " fmi2Instantiate returns null");
    }

    public void free() {
        fmiV2Interface.fmi2FreeInstance(component);
    }

    public int setDebugLogging(boolean loggingOn, String... categories) {
        return fmiV2Interface.fmi2SetDebugLogging(component, loggingOn, categories.length, categories);
    }

    public int setUpExperiment(boolean toleranceDefined, double tolerance, double startTime, boolean stopTimeDefined, double stopTime) {
        return fmiV2Interface.fmi2SetupExperiment(component, toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime);
    }

    public int enterInitializationMode() {
        return fmiV2Interface.fmi2EnterInitializationMode(component);
    }

    public int exitInitializationMode() {
        return fmiV2Interface.fmi2ExitInitializationMode(component);
    }

    public int terminate() {
        return fmiV2Interface.fmi2Terminate(component);
    }

    public int reset() {
        return fmiV2Interface.fmi2Reset(component);
    }

    public double[] getReal(int... valueReferences) {
        double[] doubleValues = new double[valueReferences.length];
        fmiV2Interface.fmi2GetReal(component, valueReferences, valueReferences.length, doubleValues);
        return doubleValues;
    }

    public int[] getInteger(int... valueReferences) {
        int[] integerValues = new int[valueReferences.length];
        fmiV2Interface.fmi2GetInteger(component, valueReferences, valueReferences.length, integerValues);
        return integerValues;
    }

    public boolean[] getBoolean(int... valueReferences) {
        int[] booleanValues = new int[valueReferences.length];
        fmiV2Interface.fmi2GetBoolean(component, valueReferences, valueReferences.length, booleanValues);
        boolean[] result = new boolean[booleanValues.length];
        for (int i = 0; i < result.length; i++) {
            result[i] = NativeBoolean.getBooleanFor((byte) booleanValues[i]);
        }
        return result;
    }

    public String[] getString(int[] valueReferences) {
        String[] stringValues = new String[valueReferences.length];
        fmiV2Interface.fmi2GetString(component, valueReferences, valueReferences.length, stringValues);
        return stringValues;
    }

    public int setReal(int[] valueReferences, double[] values) {
        return fmiV2Interface.fmi2SetReal(component, valueReferences, valueReferences.length, values);
    }

    public int setInteger(int[] valueReferences, int[] intValues) {
        return fmiV2Interface.fmi2SetInteger(component, valueReferences, valueReferences.length, intValues);
    }

    public int setBoolean(int[] valueReferences, boolean[] booleanValues) {
        int[] booleanAsIntegers = new int[booleanValues.length];
        for (int i = 0; i < booleanAsIntegers.length; i++) {
            booleanAsIntegers[i] = NativeBoolean.getValueFor(booleanValues[i]);
        }
        return fmiV2Interface.fmi2SetBoolean(component, valueReferences, valueReferences.length, booleanAsIntegers);
    }

    public int setString(int[] valueReferences, String[] stringValues) {
        return fmiV2Interface.fmi2SetString(component, valueReferences, valueReferences.length, stringValues);
    }

    public Pointer getFmuState(Pointer state) {
        PointerByReference pointerByReference = new PointerByReference(state);
        fmiV2Interface.fmi2GetFMUstate(component, pointerByReference);
        return pointerByReference.getValue();
    }

    public int setFmuState(Pointer pointer) {
        return fmiV2Interface.fmi2SetFMUstate(component, pointer);
    }

    public int freeFmuState(Pointer state) {
        return fmiV2Interface.fmi2FreeFMUstate(component, new PointerByReference(state));
    }

    public int serializedFmuStateSize(Pointer state) {
        Memory memory = new Memory(Pointer.SIZE);
        fmiV2Interface.fmi2SerializedFMUstateSize(component, state, memory);
        return memory.getInt(0);
    }

    public byte[] serializeFmuState(Pointer state) {
        int size = serializedFmuStateSize(state);
        byte[] byteBuffer = new byte[size];
        fmiV2Interface.fmi2SerializeFMUstate(component, state, byteBuffer, size);
        return byteBuffer;
    }

    public Pointer deSerializeFMUState(byte[] serializedState) {
        PointerByReference pointerByReference = new PointerByReference(new Memory(Pointer.SIZE));
        fmiV2Interface.fmi2DeSerializeFMUstate(component, serializedState, serializedState.length, pointerByReference);
        return pointerByReference.getValue();
    }

    public double[] getDirectionalDerivative(int[] unknownValueReferences, int[] knownValueReferences, double[] knownDifferential) {
        double[] sharedDoubles = new double[knownDifferential.length];
        fmiV2Interface.fmi2GetDirectionalDerivative(component, unknownValueReferences, unknownValueReferences.length, knownValueReferences, knownValueReferences.length, knownDifferential, sharedDoubles);
        return sharedDoubles;
    }

    public int setRealInputDerivatives(int[] valueReferences, int[] orders, double[] values) {
        return fmiV2Interface.fmi2SetRealInputDerivatives(component, valueReferences, valueReferences.length, orders, values);
    }

    public double[] getRealOutputDerivatives(int[] valueReferences, int[] orders) {
        double[] values = new double[valueReferences.length];
        fmiV2Interface.fmi2GetRealOutputDerivatives(component, valueReferences, valueReferences.length, orders, values);
        return values;
    }

    public int doStep(double currentSimulationTime, double stepSize, boolean noSetFmuStatePriorToCurrentPoint) {
        return fmiV2Interface.fmi2DoStep(component, currentSimulationTime, stepSize, noSetFmuStatePriorToCurrentPoint);
    }

    public int cancelStep() {
        return fmiV2Interface.fmi2CancelStep(component);
    }

    public int getStatus(int statusKind) {
        Memory status = new Memory(Pointer.SIZE);
        fmiV2Interface.fmi2GetStatus(component, statusKind, status);
        return status.getInt(0);
    }

    public double getRealStatus(int statusKind) {
        double[] realStatus = new double[1];
        fmiV2Interface.fmi2GetRealStatus(realStatus, statusKind, realStatus);
        return realStatus[0];
    }

    public int getIntegerStatus(int statusKind) {
        Memory integerStatus = new Memory(Pointer.SIZE);
        fmiV2Interface.fmi2GetIntegerStatus(integerStatus, statusKind, integerStatus);
        return integerStatus.getInt(0);
    }

    public boolean getBooleanStatus(int statusKind) {
        Memory booleanStatus = new Memory(Pointer.SIZE);
        fmiV2Interface.fmi2GetBooleanStatus(booleanStatus, statusKind, booleanStatus);
        return booleanStatus.getInt(0) == 1;
    }

    public String getStringStatus(int statusKind) {
        PointerByReference stringStatus = new PointerByReference(new Memory(Pointer.SIZE));
        fmiV2Interface.fmi2GetStringStatus(component, statusKind, stringStatus);
        return stringStatus.getValue().getString(0);
    }
}