/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.modeldescription.v2;

import org.javafmi.modeldescription.NoSuchVariableException;
import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

@Root(name = "fmiModelDescription")
public class ModelDescription implements org.javafmi.modeldescription.ModelDescription {

    @Attribute(required = true)
    private String fmiVersion;
    @Attribute(required = true)
    private String modelName;
    @Attribute(required = true)
    private String guid;
    @Attribute(required = false)
    private String description;
    @Attribute(required = false)
    private String author;
    @Attribute(required = false)
    private String version;
    @Attribute(required = false)
    private String copyright;
    @Attribute(required = false)
    private String licence;
    @Attribute(required = false)
    private String generationTool;
    @Attribute(required = false)
    private String generationDateAndTime;
    @Attribute(required = false)
    private String variableNamingConvention;
    @Attribute(required = false)
    private int numberOfEventIndicators;

    @ElementList(name = "UnitDefinitions", required = false)
    private List<Unit> unitDefinitions;
    @ElementList(name = "TypeDefinitions", required = false)
    private List<SimpleType> typeDefinitions;
    @ElementList(name = "LogCategories", required = false)
    private List<Category> logCategories;
    @Element(name = "DefaultExperiment", required = false)
    private DefaultExperiment defaultExperiment;
    @ElementList(name = "VendorAnnotations", required = false)
    private List<Tool> vendorAnnotations;
    @ElementList(name = "ModelVariables", required = true)
    private List<ScalarVariable> modelVariablesList;
    @Element(name = "ModelStructure", required = true)
    private ModelStructure modelStructure;
    @Element(name = "ModelExchange", required = false)
    private ModelExchange modelExchange;
    @Element(name = "CoSimulation", required = false)
    private CoSimulation coSimulation;

    private Map<String, ScalarVariable> modelVariables;
    private Map<Integer, String> nameReferences;
    private Capabilities capabilities;
    private String[] variablesNames;

    public ModelDescription() {
        this.modelVariablesList = new ArrayList<>();
    }

    public String getNameReferences(int vr) {
        return this.nameReferences.get(vr);
    }

    public String getModelIdentifier() {
        return coSimulation.getModelIdentifier();
    }

    public org.javafmi.modeldescription.ScalarVariable[] getModelVariables() {
        return modelVariablesList.toArray(new org.javafmi.modeldescription.ScalarVariable[modelVariablesList.size()]);
    }

    public SimpleType getTypeFromTypeDefinition(String name) {
        for(SimpleType type : typeDefinitions)
            if(type.getName().equals(name)) return type.getType();
        return null;
    }

    @Override
    public ScalarVariable getVariable(String variableName) {
        ScalarVariable scalarVariable = modelVariables.get(variableName);
        if(scalarVariable == null) throw new NoSuchVariableException(variableName);
        else return scalarVariable;
    }

    public String getModelName() {
        return modelName;
    }

    @Override
    public int[] getValueReferences(String... variablesNames) {
        int variablesValueReferences[] = new int[variablesNames.length];
        for(int i = 0; i < variablesNames.length; i++) {
            variablesValueReferences[i] = modelVariables.get(variablesNames[i]).getValueReference();
        }
        return variablesValueReferences;
    }

    @Override
    public int getValueReference(String variableName) {
        return modelVariables.get(variableName).getValueReference();
    }

    @Override
    public String[] getVariablesNames() {
        if(variablesNames != null) return variablesNames;
        variablesNames = new String[modelVariablesList.size()];
        for(int i = 0; i < variablesNames.length; i++) {
            variablesNames[i] = modelVariablesList.get(i).getName();
        }
        return variablesNames;
    }

    public String getGuid() {
        return guid;
    }

    public CoSimulation getCoSimulation() {
        return coSimulation;
    }

    public void buildAttributes() {
        modelVariables = new LinkedHashMap<>();
        for(ScalarVariable scalarVariable : modelVariablesList) {
            scalarVariable.buildAttributes();
            modelVariables.put(scalarVariable.getName(), scalarVariable);
        }
    }

    public void addVariable(ScalarVariable variable) {
        variable.setValueReference(this.nameReferences.size());
        this.nameReferences.put(this.nameReferences.size(), variable.getName());
        this.modelVariablesList.add(variable);
        this.modelVariables.put(variable.getName(), variable);

    }

    public double getDefaultTolerance() {
        return (defaultExperiment != null) ? defaultExperiment.getTolerance() : DefaultExperiment.DEFAULT_TOLERANCE;
    }

    public Capabilities getCapabilities() {
        if(capabilities == null)
            this.capabilities = buildCapabilities();
        return capabilities;
    }

    private Capabilities buildCapabilities() {
        return coSimulation.getCapabilities();
    }
}
