/*

 Copyright 2013, 2014 SIANI - ULPGC
 Jose Juan Hernandez Cabrera
 Jose Evora Gomez
 Johan Sebastian Cortes Montenegro

 This File is Part of JavaFMI Project

 JavaFMI Project is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 JavaFMI Project is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with JavaFMI Library.  If not, see <http://www.gnu.org/licenses/>.

 */
package org.javafmi.v2;

import org.javafmi.Simulation;
import org.javafmi.exceptions.JavaFmiException;
import org.javafmi.kernel.State;
import org.javafmi.kernel.Status;
import org.javafmi.kernel.v2.FmuProxy;
import org.javafmi.modeldescription.FmiVersion;
import org.javafmi.modeldescription.ModelDescription;
import org.javafmi.modeldescription.ScalarVariable;
import org.javafmi.modeldescription.v2.Capabilities;

import java.util.ArrayList;
import java.util.List;

import static java.util.Arrays.fill;

public class Access implements org.javafmi.Access {
    private final ModelDescription modelDescription;
    private FmuProxy proxy;
    private List<String> inputVariables;
    private List<String> outputVariables;

    public Access(Simulation simulation) {
        this.proxy = new ProxyRetriever(simulation).proxy;
        this.modelDescription = proxy.getModelDescription();
        buildInputAndOutputLists();
    }

    private void buildInputAndOutputLists() {
        this.inputVariables = new ArrayList<>();
        this.outputVariables = new ArrayList<>();
        for(ScalarVariable scalarVariable : modelDescription.getModelVariables()) {
            String causality = scalarVariable.getCausality();
            if(causality == null) continue;
            if(causality.equals("input"))
                inputVariables.add(scalarVariable.getName());
            else if(causality.equals("output"))
                outputVariables.add(scalarVariable.getName());
        }
    }

    @Override
    public FmiVersion getVersion() {
        return proxy.getVersion();
    }

    public String[] getKnownVariables() {
        return inputVariables.toArray(new String[inputVariables.size()]);
    }

    public String[] getUnknownVariables() {
        return outputVariables.toArray(new String[outputVariables.size()]);
    }

    public double[][] getDirectionalDerivatives(String[] unknownVariables, String[] knownVariables) {
        double[][] directionalMatrix = new double[unknownVariables.length][knownVariables.length];
        for(int i = 0; i < unknownVariables.length; i++) {
            for(int j = 0; j < knownVariables.length; j++) {
                directionalMatrix[i][j] = getDirectionalDerivative(unknownVariables[j],
                        knownVariables[i])[0];
            }
        }
        return directionalMatrix;
    }

    public double[] getDirectionalDerivative(String knownVariable, String... unknownVariables) {
        int unknownVariablesValueReferences[] = modelDescription.getValueReferences(unknownVariables);
        int knownVariablesValueReferences[] = modelDescription.getValueReferences(knownVariable);
        return proxy.getDirectionalDerivative(unknownVariablesValueReferences, knownVariablesValueReferences, new double[]{1.0});
    }

    public String getModelName() {
        return modelDescription.getModelName();
    }

    public Status doStep(double simulationTime, double stepSize) {
        return proxy.doStep(simulationTime, stepSize);
    }

    public Status cancelStep() {
        return proxy.cancelStep();
    }

    public Status terminate() {
        return proxy.terminate();
    }

    public void freeInstance() {
        proxy.freeInstance();
    }

    public Status reset() {
        return proxy.reset();
    }

    public double[] getReal(int... valueReference) {
        return proxy.getReal(valueReference);
    }

    public int getInteger(Integer valueReference) {
        return proxy.getInteger(valueReference);
    }

    public boolean getBoolean(Integer valueReference) {
        return proxy.getBoolean(valueReference);
    }

    public String getString(Integer valueReference) {
        return proxy.getString(valueReference);
    }

    public Object getEnumeration(ScalarVariable modelVariable) {
        return proxy.getEnumeration(modelVariable);
    }

    public Status setReal(int[] valueReference, double[] doubleValue) {
        return proxy.setReal(valueReference, doubleValue);
    }

    public Status setInteger(Integer valueReference, Integer intValue) {
        return proxy.setInteger(valueReference, intValue);
    }

    public Status setBoolean(Integer valueReference, Boolean booleanValue) {
        return proxy.setBoolean(valueReference, booleanValue);
    }

    public Status setString(Integer valueReference, String stringValue) {
        return proxy.setString(valueReference, stringValue);
    }

    public Status enterInitializationMode() {
        return proxy.enterInitializationMode();
    }

    public Status exitInitializationMode() {
        return proxy.exitInitializationMode();
    }

    public Status setUpExperiment(boolean toleranceDefined, double tolerance, double startTime, boolean stopTimeDefined, double stopTime) {
        return proxy.setUpExperiment(toleranceDefined, tolerance, startTime, stopTimeDefined, stopTime);
    }

    public Status setDebugLogging(boolean loggingOn, int numberOfCategories, String[] categories) {
        return proxy.setDebugLogging(loggingOn, categories);
    }

    public State getState(State state) {
        return proxy.getState(state);
    }

    public Status setState(State state) {
        return proxy.setState(state);
    }

    public Status freeState(State state) {
        return proxy.freeState(state);
    }

    public int getStateSize(State state) {
        return proxy.getStateSize(state);
    }

    public byte[] serializeState(State state) {
        return proxy.serializeState(state);
    }

    public State deserializeState(byte[] serializedState) {
        return proxy.deserializeState(serializedState);
    }

    public void instantiate(double timeout) {
        proxy.instantiate(timeout);
    }

    public Status initialize(double startTime, double stopTime) {
        return proxy.initialize(startTime, stopTime);
    }

    public Status setRealInputDerivatives(String[] realInputNames, double[] derivatives) {
        return proxy.setRealInputDerivatives(modelDescription.getValueReferences(realInputNames), getDefaultOrders(derivatives.length), derivatives);
    }

    public double[] getRealOutputDerivatives(String[] realInputNames) {
        return proxy.getRealOutputDerivatives(modelDescription.getValueReferences(realInputNames), getDefaultOrders(realInputNames.length));
    }

    private int[] getDefaultOrders(int length) {
        int[] orders = new int[length];
        fill(orders, 1);
        return orders;
    }

    public int[] getValueReferences(String[] variables) {
        return modelDescription.getValueReferences(variables);
    }

    public int getValueReference(String variable) {
        return modelDescription.getValueReference(variable);
    }

    public Capabilities getCapabilities() {
        return proxy.getCapabilities();
    }

    private class ProxyRetriever extends Simulation {
        private final FmuProxy proxy;

        public ProxyRetriever(Simulation simulation) {
            checkVersion(simulation);
            this.proxy = (FmuProxy) getProxy(simulation);
        }

        private void checkVersion(Simulation simulation) {
            if(getProxy(simulation) instanceof FmuProxy) return;
            throw new JavaFmiException("Access to FMU does not match with Simulation version");
        }
    }
}
